// $Id$
// Projekt Name: InfoOS
// Licensed unter the GPL (General Public License)
// Written by Sebastian Tilders

// pmm (Physical Memory Manager)
// Diese Datei ist für die Physikalische Speicherverwaltung zuständig. Dazu wird eine Bitmap erstellt
// die Anzeigt ob Speicherbereiche frei (1) oder Belegt (0) sind. Jedes Bit steht für eine 4kb Page (Paging).
// Wir gehen von Maximal 4 GB (4GiB / 4KiB = 1MiB großes Array) aus. 1MiB / 32 Bit = 32768 Einträge
#include "../header/typedefs.h"
#include"../header/multiboot.h"
#include "../header/display.h"
#include "../header/string.h"
#define BIT_POSITION (i * 32) + (i2)
/*
(i * 32 + i2) * 4096 = PhysAddr
(32i + i2)*4096 = PhysAddr
131072i + 4096i2 = PhysAddr
- 131072i
4096i2 = PhysAddr - 131072i
/ 4096
i2 = (PhysAddr - 131072i) / 4096
*/
#define PHYSIKAL_ADDR BIT_POSITION * 4096
#define ALL_BITMAP_ADDR int_byte_addr / 4096
// (32Bit * 4096 Bytes) = 131072 Bytes = Speichergröße eines Eintrags 
#define I_BITMAP_ADDR (int_byte_addr / 131072)
#define I2_BITMAP_ADDR (int_byte_addr/4096) - (I_BITMAP_ADDR * 32)
static uint32_t pmm_bitmap[32768];

// Alloziert eine 4k Page im Speicher. Wenn kein freier Speicher mehr vorhanden ist, gibt es einen Zeiger auf 0x0 aus.
uint32_t* pmm_alloc()
{
  uint16_t i, i2;
   for(i = 0; i < 32768; i++)
   {
     for(i2 = 0; i2<32; i2++)
     {
       if((pmm_bitmap[i] & (1 << i2)) > 0)
       {
		pmm_bitmap[i] &= ~(1<<i2);
		return (void*) PHYSIKAL_ADDR;
       }
     }
   }
   return (uint32_t*) 0x0;
}

// Gibt eine 4k Page im Speicher wieder frei
void pmm_free(uint8_t* byte_addr)
{
  uint32_t int_byte_addr = (uint32_t) byte_addr;
  int_byte_addr &= ~0xFFF;	// Abrunden auf 4k Adresse
  uint32_t i = I_BITMAP_ADDR;
  uint32_t i2 = I2_BITMAP_ADDR;
  pmm_bitmap[i] |= (1 << i2);
  return;
}
void pmm_mark_used(uint8_t* byte_addr)
{
	uint32_t int_byte_addr = (uint32_t) byte_addr;
	int_byte_addr &= ~0xFFF;
	uint32_t i = I_BITMAP_ADDR;
	uint32_t i2 = I2_BITMAP_ADDR;
	pmm_bitmap[i] &= ~(0x1 << i2);
	return;
}
// Initalisiert die Speicherverwaltung, sprich belegt volle Berreiche in der Bitmap.
/*
 * Die Speicherverwaltung sucht sich aus der Multibootstruktur die passenden Einträge für die
 * vorbelegung der Bits in die Bitmap heraus. Dazu zählen Bootloader, geladene Module und der 
 * Kernel.
 */
void pmm_init(struct multiboot* mbs_info)
{
	extern const void kernel_start;
	extern const void kernel_ende;
	uint8_t* addr;
	struct mbs_mods* modules = (struct mbs_mods*) mbs_info->mbs_mods_addr;
	pmm_mark_used((uint8_t*) modules);
	pmm_mark_used((uint8_t*) mbs_info);
	memset((char*) pmm_bitmap,0,sizeof(pmm_bitmap));
	struct mbs_mmap* mmap = (void*) mbs_info->mbs_mmap_addr;
	struct mbs_mmap* mmap_end = (void*) ((uint8_t*) mbs_info->mbs_mmap_addr + mbs_info->mbs_mmap_length);
	while(mmap < mmap_end)
	{
		if(mmap->Type == 1)
		{
			uint8_t* map_start = mmap->BaseAddr;
			uint8_t* map_end = map_start + mmap->Length;
			while(map_start < map_end)
			{
				pmm_free((void*)map_start);
				map_start += 0x1000;
			}
		}
		mmap++;
	}
	uint8_t* kernel_ptr = (uint8_t*) &kernel_start;
	while(kernel_ptr < (uint8_t*) &kernel_ende)
	{
		pmm_mark_used(kernel_ptr); kernel_ptr += 0x1000;
	}
	pmm_mark_used((uint8_t*) 0x200000);
	pmm_mark_used((void*) 0x0);
	for(uint8_t i = 0; i < mbs_info->mbs_mods_count; i++)
	{
		addr = modules[i].mod_start;
		while(addr < (uint8_t*) modules[i].mod_end)
		{
			pmm_mark_used(addr);
			addr += 0x1000;
		}
	}
return;
}
uint32_t free_memory(void) 
{
	uint32_t free_memory_var = 0x0;
	uint32_t i, i2;
	for(i = 0;i < 32768;i++)
	{
			for(i2=0;i2<32;i2++)
			{
				if((uint32_t)((pmm_bitmap[i]>>i2) & 1) == 1)
				{
					free_memory_var += 0x1000;
				}
			}
	}
	return free_memory_var;
}
